from snakebids.utils.output import (
    Mode,
    get_time_hash,
    prepare_output,
    retrofit_output,
    write_config_file,
    write_output_mode,
)
from snakebids.utils.snakemake_io import (
    glob_wildcards,
    regex,
    update_wildcard_constraints,
)

__all__ = [
    "Mode",
    "get_time_hash",
    "glob_wildcards",
    "prepare_output",
    "regex",
    "retrofit_output",
    "update_wildcard_constraints",
    "write_config_file",
    "write_output_mode",
]



"""Tools to manage generation and interconversion of bidsapps and snakemake outputs."""

import hashlib
import itertools as it
import json
import os
import shutil
import time
from collections import OrderedDict
from pathlib import Path, PosixPath, WindowsPath
from typing import Iterable, List, Union

import yaml
from colorama import Fore
from progress.bar import IncrementalBar
from typing_extensions import Literal

from snakebids.exceptions import RunError

Mode = Union[Literal["workflow"], Literal["bidsapp"]]


def prepare_output(
    src: Path, outputdir: Path, mode: Mode, force_conversion: bool = False
):
    """Ensure output directory is in the correct mode and is ready for snakemake to run

    Checks for existing output at the directory, converting to the appropriate mode if
    necessary. If running in workflow mode, the src snakemake directory is copied over
    to the output, and snakemake output will be put in results. In bidsapp mode, the
    snakemake results will be put in the top level of the output directory without any
    extra snakemake files. Creates a .snakebids file to track output mode and other
    workflow information. Raises exceptions when attempting to convert from workflow
    mode to bidsapp mode without force, when the output directory already has contents
    but no .snakebids file to identify it, or when attempting to convert snakemake
    results that, themselves, have a results folder

    Parameters
    ----------
    src : Path
        Path to snakebids app
    outputdir : Path
        Path to output
    mode : "bidsapp" or "workflow"
        Mode in which to run
    force_conversion : bool
        Force conversion from workflow to bidsapp
        mode. Defaults to False.

    Returns
    -------
    Path
        Path to new root folder (output for bidsapp, output/results for workflow)

    Raises
    ------
    RunError
        Raised when attempting to convert from workflow mode to bidsapp mode
        without force, when the output directory already has contents but no
        .snakebids file to identify it, or when attempting to convert snakemake
        results that, themselves, have a results folder

    """
    if mode not in ["workflow", "bidsapp"]:
        raise RunError(
            f"Requested unsupported output mode: {mode}.\n"
            'Please select between "workflow" and "bidsapp"'
        )

    # Look for .snakebids file. If the outputdir doesn't yet exist, we'll get None.
    # If it does exist but there's no .snakebids file, an error will be raised.
    snakebids_file = _get_snakebids_file(outputdir)
    outputdir.mkdir(exist_ok=True)

    if not snakebids_file:
        write_output_mode(outputdir / ".snakebids", mode)

        if mode == "bidsapp":
            return outputdir

        return _copy_snakemake_app(src, outputdir)

    with snakebids_file.open("r") as f:
        snakebids_data = json.load(f)
    root = _convert_output(
        snakebids_data["mode"], mode, src, outputdir, force_conversion
    )
    write_output_mode(outputdir / ".snakebids", mode)
    return root


def write_output_mode(dotfile: Path, mode: Mode):
    """Write output mode to .snakebids

    Parameters
    ----------
    dotfile: Path
        Path to .snakebids file to be written
    mode: Mode
        Mode to write: either "bidsapp" or "workflow"
    """
    if dotfile.exists():
        with dotfile.open("r") as f:
            data = json.load(f)
    else:
        data = {}
    data["mode"] = mode
    with (dotfile).open("w") as f:
        json.dump(data, f)


def retrofit_output(output: Path, config_files: Iterable[Path]):
    """Convert legacy snakebids output to bidsapp mode.

    Expects a directory containing previous snakebids outpus. This should contain one or
    more config files, as specified in the config_files parameter. If a config directory
    resides in the output, it should only contain specified config files. If extra files
    are found, an error will be thrown. All config files will be deleted, and a new
    .snakebids file will be created.

    Parameters
    ----------
    output : Path
        Path of output directory to be modified
    config_files : Iterable[Path]
        List of paths of config files.

    Returns
    -------
    boolean
        Returns True if successful, False if User declined operation in interactive
        prompt

    Raises
    ------
    RunError
        Raised if existing .snakebids file found, if unrecognized files are found in the
        config folder, or if no config files are provided.
    """
    config_files = [*config_files]
    if (output / ".snakebids").exists():
        raise RunError(f".snakebids file already found at {output}")
    if not config_files:
        raise RunError(f"No config files found in {output}. Cannot perform retrofit.")
    if (output / "config").exists() and (output / "config").is_dir():
        to_delete = it.chain(config_files, [output / "config"])
        unknown_files = {*(output / "config").iterdir()} - {*config_files}
        if len(unknown_files) > 0:
            raise RunError(
                f"Unrecognized files found in config folder ({output/'config'}",
                _format_path_list(unknown_files),
            )
    else:
        to_delete = config_files

    print(f"Converting {output} into bidsapp format.\n")

    if not _remove_all(to_delete, confirm=True):
        return False
    write_output_mode(output / ".snakebids", "bidsapp")
    return True


def _convert_output(start: Mode, end: Mode, src: Path, output: Path, force=False):
    """Convert existing output between bidsapp and workflow mode

    Does nothing if start and end are the same. Because of the potential loss of
    information (all the workflow files), this will only convert from bidsapp to
    workflow mode if force is True, otherwise it raises an exception. Returns the
    root folder of the converted dataset (ouput if bidsapp mode, results if workflow)

    Parameters
    ----------
    start : "bidsapp" or "workflow"
        Current format of the output
    end : "bidsapp" or "workflow"
        Desired format of the output
    src : Path
        Path to the snakebids app being run
    output : Path
        Output to transform
    force : bool
        Force conversions from workflow to bidsapp mode.
        Defaults to False.

    Returns
    -------
    Path
        Path to the root folder (output/results if workflow, output if bidsapp)

    Raises
    ------
    RunError
        Raised if conversion from workflow to bidsapp attempted without force

    """
    if start == end:
        if end == "workflow":
            return output / "results"
        return output

    # Convert to workflow mode
    if end == "workflow":
        print(
            f"{Fore.GREEN}Found bidsapp output at {Fore.RESET}{output.resolve()}\n\n"
            f"{Fore.GREEN}Converting to workflow mode...{Fore.RESET}\n"
        )
        results = _check_for_results_folder(output)
        results.mkdir()
        for f in output.iterdir():
            if f != results:
                shutil.move(f, results / f.name)

        # Move .snakebids file back to the top level
        shutil.move(results / ".snakebids", output / ".snakebids")
        _copy_snakemake_app(src, output, False)
        return results

    # Convert to Bidsapp mode
    if not force:
        raise RunError(
            f"You are attempting to convert a preexisting output ({output}) "
            "from workflow mode to bidsapp mode. This will result in the loss "
            "of all workflow files and custom configs. If you are sure you "
            "wish to do this, run this command again with --force-conversion. "
            "Otherwise, run the command with --workflow-mode to maintain the current "
            "output mode."
        )
    # Check if results folder/file exists within the results folder. We don't
    # need its output, just its exception.
    _check_for_results_folder(output / "results")

    # Delete everything in the output folder except for .snakebids and results
    _remove_all(
        f
        for f in output.iterdir()
        if f not in [output / ".snakebids", output / "results", output / ".snakemake"]
    )

    for f in (output / "results").iterdir():
        shutil.move(f, output / f.name)

    (output / "results").rmdir()
    return output


def _remove_all(paths: Iterable[Path], confirm: bool = False):
    if confirm:
        paths = [*paths]
        print(
            f"\t{Fore.YELLOW}The following files and folders will be DELETED:\n"
            f"{Fore.RESET}",
            _format_path_list(paths),
        )
        user_response = input("Would you like to continue? [yes,NO]")
        if user_response.lower() != "yes":
            return False

    dirs: List[Path] = []
    for path in paths:
        if path.is_dir():
            dirs.append(path)
        if path.is_symlink():
            path.unlink()
        if path.is_file():
            os.remove(path)
    for d in dirs:
        shutil.rmtree(d)
    return True


def _format_path_list(paths: Iterable[Path]):
    return "\t\t- " + "\n\t\t- ".join(str(p) for p in paths)


def _copy_snakemake_app(src: Path, dest: Path, create_results: bool = True):
    """Copies snakemake app from src to dest, skipping config and results directories

    Creates an empty results and config folder, and returns the path to the results
    folder.

    Parameters
    ----------
    src : Path
        Directory to copy from
    dest : Path
        Directory to copy to
    create_results: bool
        If True, create an empty results folder (Default value = True)

    Returns
    -------
    Path
        Path of results folder in new snakemake location

    """
    # shutils.copytree makes it hard to exclude just root level folders, so we do this
    # manually
    file_list = [
        *it.chain(
            # All root level items
            f.relative_to(src)
            for f in Path(src).iterdir()
            if f
            not in [
                src / ".snakebids",
                src / "config",
                src / "results",
                src / ".snakemake",
            ]
        )
    ]

    # Iterate through files
    progress_bar = IncrementalBar("Copying snakebids app", max=len(file_list))
    for file in file_list:
        if (src / file).is_dir():
            shutil.copytree(src / file, dest / file)
        # Copy over files, making parents as necessary
        elif (src / file).is_file():
            (dest / file).parent.mkdir(parents=True, exist_ok=True)
            shutil.copy(src / file, dest / file, follow_symlinks=False)
        progress_bar.next()

    if create_results:
        (dest / "results").mkdir()
    (dest / "config").mkdir()
    return dest / "results"


def _get_snakebids_file(outputdir: Path):
    """Ensure populated dir contains .snakebids file, retrieving it if it does.

    First checks if outputdir doesn't exist or is completely empty, returing None if so.
    If it does have data, it checks for a .snakebids file, returning its Path if found.
    If no .snakebids file is found, it raises an exception.

    Parameters
    ----------
    outputdir : Path
        Directory to check.

    Returns
    -------
    Path or None
        None if output dir is nonexistant or empty, otherwise the path
        of the .snakebids file

    Raises
    ------
    RunError
        Raised if outputdir contains contents but no .snakebids file

    """
    # Check if outputdir exits
    if not outputdir.exists():
        return None

    # If it does exist, is it empty?
    if len([*outputdir.iterdir()]) == 0:
        return None

    # If it's not empty, is there a .snakebids file?
    if (outputdir / ".snakebids").exists():
        return outputdir / ".snakebids"

    # We have an occupied directory without a .snakebids file, so we have no idea
    # what's there.
    raise RunError(
        f"Output dir `{outputdir.resolve()}` exists, but `.snakebids` file "
        "not found. Please specify either a new directory, or a ",
        "directory where you've previously run this Snakebids app.",
    )


def _check_for_results_folder(root: Path):
    """Check folder for results folder

    Raises exception if it does exist, otherwise returns the name of the folder it
    looked for.

    Parameters
    ----------
    root : Path
        Folder in which to search

    Returns
    -------
    Path
        Name of folder searched for.

    Raises
    ------
    RunError
        Raised if results folder found

    """
    results = root / "results"
    if results.exists():
        raise RunError(
            "Cannot convert output format as a results folder or file already "
            f"exists in your data directory ({results}). Please rename or remove "
            "this item."
        )
    return results


def get_time_hash():
    """currently unused"""

    time_hash = hashlib.sha1()
    time_hash.update(str(time.time()).encode("utf-8"))
    return time_hash.hexdigest()[:8]


def write_config_file(config_file: Path, data: dict, force_overwrite: bool = False):
    if (config_file.exists()) and not force_overwrite:
        raise RunError(
            f"A config file named {config_file.name} already exists:\n"
            f"\t- {config_file.resolve()}\n"
            "Please move or rename either the existing or incoming config."
        )
    config_file.parent.mkdir(exist_ok=True)

    # TODO: copy to a time-hashed file for provenance purposes?
    #       unused as of now..
    # time_hash = get_time_hash()

    with open(config_file, "w", encoding="utf-8") as f:
        # write either as JSON or YAML
        if config_file.suffix == ".json":
            json.dump(data, f, indent=4)
            return

        # if not json, then should be yaml or yml

        # this is needed to make the output yaml clean
        yaml.add_representer(
            OrderedDict,
            lambda dumper, data: dumper.represent_mapping(
                "tag:yaml.org,2002:map", data.items()
            ),
        )

        # Represent any PathLikes as str.
        def path2str(dumper, data):
            return dumper.represent_scalar("tag:yaml.org,2002:str", str(data))

        yaml.add_representer(PosixPath, path2str)
        yaml.add_representer(WindowsPath, path2str)

        yaml.dump(data, f, default_flow_style=False, sort_keys=False)



"""Utilities for converting Snakemake apps to BIDS apps."""

import json
import logging
import os
import re
from collections import OrderedDict
from os.path import join
from pathlib import Path

from bids import BIDSLayout, BIDSLayoutIndexer

from snakebids.utils.snakemake_io import glob_wildcards

_logger = logging.getLogger(__name__)


# pylint: disable=too-many-arguments
def bids(
    root=None,
    datatype=None,
    prefix=None,
    suffix=None,
    subject=None,
    session=None,
    include_subject_dir=True,
    include_session_dir=True,
    **entities,
):
    """Helper function for generating bids paths for snakemake workflows.

    File path is of the form::

        [root]/[sub-{subject}]/[ses-{session]/
        [prefix]_[sub-{subject}]_[ses-{session}]_[{key}-{val}_ ... ]_[suffix]

    Parameters
    ----------
    root : str or Path, default=None
        root folder to include in the path (e.g. 'results')
    datatype : str, default=None
        folder to include after sub-/ses- (e.g. anat, dwi )
    prefix : str, default=None
        string to prepend to the file name (typically not defined, unless you
        want tpl-{tpl}, or a datatype)
    suffix : str, default=None
        bids suffix including extension (e.g. 'T1w.nii.gz')
    subject : str, default=None
        subject to use, for folder and filename
    session : str, default=None
        session to use, for folder and filename
    include_subject_dir : bool, default=True
        whether to include the sub-{subject} folder if subject defined
        (default: True)
    include_session_dir : bool, default=True
        whether to include the ses-{session} folder if session defined
        (default: True)
    **entities : dict, optional
        dictionary of bids entities (e.g. space=T1w for space-T1w)

    Returns
    -------
    str
        bids-like file path

    Examples
    --------

    Below is a rule using bids naming for input and output::

        rule proc_img:
            input: 'sub-{subject}_T1w.nii.gz'
            output: 'sub-{subject}_space-snsx32_desc-preproc_T1w.nii.gz'

    With bids() you can instead use::

         rule proc_img:
            input: bids(subject='{subject}',suffix='T1w.nii.gz')
            output: bids(
                subject='{subject}',
                space='snsx32',
                desc='preproc',
                suffix='T1w.nii.gz'
            )

    Note that here we are not actually using "functions as inputs" in
    snakemake, which would require a function definition with wildcards as
    the argument, and restrict to input/params, but bids() is being used
    simply to return a string.

    Also note that space, desc and suffix are NOT wildcards here, only
    {subject} is. This makes it easy to combine wildcards and non-wildcards
    with bids-like naming.

    However, you can still use bids() in a lambda function. This is
    especially useful if your wildcards are named the same as bids entities
    (e.g. {subject}, {session}, {task} etc..)::

        rule proc_img:
            input: lambda wildcards: bids(**wildcards,suffix='T1w.nii.gz')
            output: bids(
                subject='{subject}',
                space='snsx32',
                desc='preproc',
                suffix='T1w.nii.gz'
            )

    Or another example where you may have many bids-like wildcards used in
    your workflow::

        rule denoise_func:
            input: lambda wildcards: bids(**wildcards, suffix='bold.nii.gz')
            output: bids(
                subject='{subject}',
                session='{session}',
                task='{task}',
                acq='{acq}',
                desc='denoise',
                suffix='bold.nii.gz'
            )

    In this example, all the wildcards will be determined from the output
    and passed on to bids() for inputs. The output filename will have a
    'desc-denoise' flag added to it.

    Also note that even if you supply entities in a different order, the
    entities will be ordered based on the OrderedDict defined here.
    If entities not known are provided, they will be just be placed
    at the end (before the suffix), in the order you provide them in.

    Notes
    -----

    * For maximum flexibility all arguments are optional (if none are
      specified, will return empty string)

    * Some code adapted from mne-bids, specifically
      https://mne.tools/mne-bids/stable/_modules/mne_bids/utils.html
    """

    # replace underscores in keys (needed to that users can use reserved
    # keywords by appending a _)
    entities = {k.replace("_", ""): v for k, v in entities.items()}

    # strict ordering of bids entities is specified here:
    order = OrderedDict(
        [
            ("task", None),
            ("acq", None),
            ("ce", None),
            ("rec", None),
            ("dir", None),
            ("run", None),
            ("mod", None),
            ("echo", None),
            ("hemi", None),
            ("space", None),
            ("res", None),
            ("den", None),
            ("label", None),
            ("desc", None),
        ]
    )

    # Now add in entities (this preserves ordering above)
    for key, val in entities.items():
        order[key] = val

    # initialize lists for filename and folder
    # will append to these, then '_'.join() os.path.join() respectively
    filename = []
    folder = []

    # root directory
    if isinstance(root, str):
        folder.append(root)
    elif isinstance(root, Path):
        folder.append(str(root.resolve()))

    # if prefix is defined, put it before other anything else
    if isinstance(prefix, str):
        filename.append(prefix)

    # if subject defined then append to file and folder
    if isinstance(subject, str):
        if include_subject_dir is True:
            folder.append(f"sub-{subject}")
        filename.append(f"sub-{subject}")

    # if session defined then append to file and folder
    if isinstance(session, str):
        if include_session_dir is True:
            folder.append(f"ses-{session}")
        filename.append(f"ses-{session}")

    if isinstance(datatype, str):
        folder.append(datatype)

    # add the entities
    filename += [f"{key}-{val}" for key, val in order.items() if val is not None]

    # if suffix is defined, append it
    if isinstance(suffix, str):
        filename.append(suffix)

    if len(filename) == 0:
        return ""

    # now, join up the lists:
    filename = "_".join(filename)

    if len(folder) > 0:
        filename = join(*folder, filename)

    return filename


def print_boilerplate():
    """Function to print out boilerplate to add to Snakefile. (not used
    anywhere yet)"""

    print(
        """
# ---- begin snakebids boilerplate ------------------------------------------

import snakebids
from snakebids import bids

configfile: 'config/snakebids.yml'

#writes inputs_config.yml and updates config dict
config.update(snakebids.generate_inputs(bids_dir=config['bids_dir'],
                            pybids_inputs=config['pybids_inputs'],
                            derivatives=config['derivatives'],
                            participant_label=config['participant_label'],
                            exclude_participant_label=config['exclude_participant_label']))


#this adds constraints to the bids naming
wildcard_constraints:  **snakebids.get_wildcard_constraints(
    config['pybids_inputs']
)

# ---- end snakebids boilerplate --------------------------------------------
"""
    )


def filter_list(zip_list, wildcards, return_indices_only=False):
    """This function is used when you are expanding over some subset of the
    wildcards i.e. if your output file doesn't contain all the wildcards in
    input_wildcards

    Parameters
    ----------
    zip_list : dict
        generated zip lists dict from config file to filter

    wildcards : dict
        wildcard values to filter the zip lists

    return_indices_only : bool, default=False
        return the indices of the matching wildcards

    Returns
    -------
    dict
        zip list with non-matching elements removed

    Examples
    --------
    ::

        >>> import snakebids

    Filtering to get all ``subject='01'`` scans::

        >>> snakebids.filter_list(
        ...     {
        ...         'dir': ['AP','PA','AP','PA', 'AP','PA','AP','PA'],
        ...         'acq': ['98','98','98','98','99','99','99','99'],
        ...         'subject': ['01','01','02','02','01','01','02','02' ]
        ...     },
        ...     {'subject': '01'}
        ... ) == {
        ...     'dir': ['AP', 'PA', 'AP', 'PA'],
        ...     'acq': ['98', '98', '99', '99'],
        ...     'subject': ['01', '01', '01', '01']
        ... }
        True

    Filtering to get all ``acq='98'`` scans::

        >>> snakebids.filter_list(
        ...     {
        ...         'dir': ['AP','PA','AP','PA', 'AP','PA','AP','PA'],
        ...         'acq': ['98','98','98','98','99','99','99','99'],
        ...         'subject': ['01','01','02','02','01','01','02','02' ]
        ...     },
        ...     {'acq': '98'}
        ... ) == {
        ...     'dir': ['AP', 'PA', 'AP', 'PA'],
        ...     'acq': ['98', '98', '98', '98'],
        ...     'subject': ['01', '01', '02', '02']
        ... }
        True

    Filtering to get all ``dir=='AP'`` scans::

        >>> snakebids.filter_list(
        ...     {
        ...         'dir': ['AP','PA','AP','PA', 'AP','PA','AP','PA'],
        ...         'acq': ['98','98','98','98','99','99','99','99'],
        ...         'subject': ['01','01','02','02','01','01','02','02' ]
        ...     },
        ...     {'dir': 'AP'}
        ... ) == {
        ...     'dir': ['AP', 'AP', 'AP', 'AP'],
        ...     'acq': ['98', '98', '99', '99'],
        ...     'subject': ['01', '02', '01', '02']
        ... }
        True

    Filtering to get all ``subject='03'`` scans (i.e. no matches)::

        >>> snakebids.filter_list(
        ...     {
        ...         'dir': ['AP','PA','AP','PA', 'AP','PA','AP','PA'],
        ...         'acq': ['98','98','98','98','99','99','99','99'],
        ...         'subject': ['01','01','02','02','01','01','02','02' ]
        ...     },
        ...     {'subject': '03'}
        ... ) == {
        ...     'dir': [],
        ...     'acq': [],
        ...     'subject': []
        ... }
        True
    """
    keep_indices = []
    for key, val in wildcards.items():
        # get indices where wildcard exists
        if key not in zip_list.keys():
            continue
        indices = [i for i, v in enumerate(zip_list[key]) if v == val]
        if len(keep_indices) == 0:
            keep_indices = indices
        else:
            keep_indices = [x for x in indices if x in keep_indices]
    # Now we have the indices, so filter the lists
    if return_indices_only:
        return keep_indices
    return {
        key: [zip_list[key][i] for i in keep_indices] for key, val in zip_list.items()
    }


def get_filtered_ziplist_index(zip_list, wildcards, subj_wildcards):
    """Use this function when you have wildcards for a single scan instance,
    and want to know the index of that scan, amongst that subject's scan
    instances.

    Parameters
    ----------
    zip_list : dict
        lists for scans in a dataset, zipped to get each instance
    wildcards: dict
        wildcards for the single instance for querying it's index
    subj_wildcards: dict
        keys of this dictionary are used to pick out the subject/(session)
        from the wildcards

    Examples
    --------

    >>> import snakebids

    In this example, we have a dataset where with scans from two subjects,
    where each subject has ``dir-AP`` and ``dir-PA`` scans, along with
    ``acq-98`` and ``acq-99``:

    * ``sub-01_acq-98_dir-AP_dwi.nii.gz``
    * ``sub-01_acq-98_dir-PA_dwi.nii.gz``
    * ``sub-01_acq-99_dir-AP_dwi.nii.gz``
    * ``sub-01_acq-99_dir-PA_dwi.nii.gz``
    * ``sub-02_acq-98_dir-AP_dwi.nii.gz``
    * ``sub-02_acq-98_dir-PA_dwi.nii.gz``
    * ``sub-02_acq-99_dir-AP_dwi.nii.gz``
    * ``sub-02_acq-99_dir-PA_dwi.nii.gz``

    The ``zip_list`` produced by ``generate_inputs()`` is the set of entities
    that when zipped together, e.g. with ``expand(path, zip, **zip_list)``,
    produces the entity combinations that refer to each scan::

        {
            'dir': ['AP','PA','AP','PA', 'AP','PA','AP','PA'],
            'acq': ['98','98','98','98','99','99','99','99'],
            'subject': ['01','01','02','02','01','01','02','02']
        }

    The ``filter_list()`` function produces a subset of the entity
    combinations as a filtered zip list. This is used e.g. to get all the
    scans for a single subject.

    This ``get_filtered_ziplist_index()`` function performs ``filter_list()``
    twice:

    1. Using the ``subj_wildcards`` (e.g.: ``'subject': '{subject}'``) to get
       a subject/session-specific ``zip_list``.
    2. To return the indices from that list of the matching wildcards.

    In this example, if the wildcards parameter was::

        {'dir': 'PA', 'acq': '99', 'subject': '01'}

    Then the first (subject/session-specific) filtered list provides this zip
    list::

        {
            'dir': ['AP','PA','AP','PA'],
            'acq': ['98','98','99','99'],
            'subject': ['01','01','01','01']
        }

    which has 4 combinations, and thus are indexed from 0 to 3.

    The returned value would then be the index (or indices) that matches the
    wildcards. In this case, since the wildcards were
    ``{'dir': 'PA', 'acq': '99', 'subject':'01'}``, the return index is 3. ::

        >>> snakebids.get_filtered_ziplist_index(
        ...     {
        ...         'dir': ['AP','PA','AP','PA', 'AP','PA','AP','PA'],
        ...         'acq': ['98','98','98','98','99','99','99','99'],
        ...         'subject': ['01','01','02','02','01','01','02','02' ]
        ...     },
        ...     {'dir': 'PA', 'acq': '99', 'subject': '01'},
        ...     {'subject': '{subject}' }
        ... )
        3
    """
    # get the subject/(session) dict:
    subj_dict = {key: wildcards[key] for key in subj_wildcards.keys()}

    # now filter the list based on subj_wildcards
    zip_list_filtered = filter_list(zip_list, subj_dict)

    # get the index of the wildcard from this filtered list
    indices = filter_list(zip_list_filtered, wildcards, return_indices_only=True)
    if len(indices) == 1:
        return indices[0]
    return indices


def _read_bids_tags(bids_json=None):
    """Read the bids tags we are aware of from a JSON file. This is used
    specifically for compatibility with pybids, since some tag keys are
    different from how they appear in the file name, e.g. ``subject`` for
    ``sub``, and ``acquisition`` for ``acq``.

    Parameters
    ----------
    bids_json : str, optional
        Path to JSON file to use, if not specified will use
        ``bids_tags.json`` in the snakebids module.

    Returns
    -------
    dict:
        Dictionary of bids tags"""
    if bids_json is None:
        bids_json = Path(__file__).parent.resolve() / "bids_tags.json"
    with bids_json.open("r") as infile:
        bids_tags = json.load(infile)
    return bids_tags


def _generate_search_terms(participant_label=None, exclude_participant_label=None):
    search_terms = {}

    if participant_label is not None and exclude_participant_label is not None:
        raise ValueError(
            "Cannot define both participant_label and "
            "exclude_participant_label at the same time"
        )

    # add participant_label or exclude_participant_label to search terms (if
    # defined)
    # we make the subject key in search_terms a list so we can have both
    # participant_label and exclude_participant_label defined
    if participant_label is not None:
        if isinstance(participant_label, list):
            search_terms["subject"] = participant_label
        else:
            search_terms["subject"] = [participant_label]

    if exclude_participant_label is not None:
        # if multiple subjects to exclude, combine with with subj1|subj2|...
        if isinstance(exclude_participant_label, list):
            exclude_string = "|".join(exclude_participant_label)
        # if not, then string is the label itself
        else:
            exclude_string = exclude_participant_label
        search_terms["regex_search"] = True
        # regex to exclude subjects
        search_terms["subject"] = [f"^((?!({exclude_string})).)*$"]
    return search_terms


def generate_inputs(
    bids_dir,
    pybids_inputs,
    derivatives=False,
    search_terms=None,
    limit_to=None,
    participant_label=None,
    exclude_participant_label=None,
):
    """Dynamically generate snakemake inputs using pybids_inputs dict, and
    pybids to parse the bids dataset.

    Parameters
    ----------
    bids_dir : str
        Path to bids directory

    pybids_inputs : dict
        Configuration for bids inputs, with keys as the names (``str``)

        Nested `dicts` with the following required keys:

        * ``"filters"``: Dictionary containing keyword arguments that will
          be passed to pybids ``get()``.

        * ``"wildcards"``: List of (str) bids tags to include as wildcards in
          snakemake. At minimum this should usually include
          ``['subject','session']``, plus any other wildcards that you may
          want to make use of in your snakemake workflow, or want to retain
          in the output paths. Any wildcards in this list that are not in the
          filename will just be ignored.

    Returns
    -------
    dict:
        The dict returned by this functions contains seven items. Each of
        the following four items is a dict containing one item for each
        modality described by ``pybids_inputs``.

        * ``"input_path"``: String with a wildcard-filled path that matches
          the images for this modality.

        * ``"input_zip_lists"``: Dictionary where each key is a wildcard
          entity and each value is a list of the values found for that
          entity. Each of these lists has length equal to the number of
          images matched for this modality, so they can be zipped together to
          get a list of the wildcard values for each image.

        * ``"input_lists"``: Dictionary where each key is a wildcard entity
          and each value is a list of the unique values found for that
          entity. These lists may not be the same length.

        * ``"input_wildcards"``: Dictionary where each key is the name of a
          wildcard entity, and each value is the Snakemake wildcard used for
          that entity.

        Then there are three more top-level entries in the dictionary:

        * ``"subjects"``: A list of the subjects in the dataset.

        * ``"sessions"``: A list of the sessions in the dataset.

        * ``"subj_wildcards"``: The subject and session wildcards applicable
          to this dataset. ``{"subject": "{subject}"}`` if there is only one
          session, ``{"subject": "{subject}", "session": "{session}"}`` if
          there are multiple sessions.

    Notes
    -----
    As an example, consider the following BIDS dataset::

        bids-example/
        ├── dataset_description.json
        ├── participants.tsv
        ├── README
        └── sub-control01
            ├── anat
            │   ├── sub-control01_T1w.json
            │   ├── sub-control01_T1w.nii.gz
            │   ├── sub-control01_T2w.json
            │   └── sub-control01_T2w.nii.gz
            ├── dwi
            │   ├── sub-control01_dwi.bval
            │   ├── sub-control01_dwi.bvec
            │   └── sub-control01_dwi.nii.gz
            ├── fmap
            │   ├── sub-control01_magnitude1.nii.gz
            │   ├── sub-control01_phasediff.json
            │   ├── sub-control01_phasediff.nii.gz
            │   └── sub-control01_scans.tsv
            └── func
                ├── sub-control01_task-nback_bold.json
                ├── sub-control01_task-nback_bold.nii.gz
                ├── sub-control01_task-nback_events.tsv
                ├── sub-control01_task-nback_physio.json
                ├── sub-control01_task-nback_physio.tsv.gz
                ├── sub-control01_task-nback_sbref.nii.gz
                ├── sub-control01_task-rest_bold.json
                ├── sub-control01_task-rest_bold.nii.gz
                ├── sub-control01_task-rest_physio.json
                └── sub-control01_task-rest_physio.tsv.gz

    With the following ``pybids_inputs`` defined in the config file::

        pybids_inputs:
          bold:
            filters:
              suffix: 'bold'
              extension: '.nii.gz'
              datatype: 'func'
            wildcards:
              - subject
              - session
              - acquisition
              - task
              - run

    Then ``generate_inputs(bids_dir, pybids_input)`` would return the
    following dictionary::

        {
            "input_path": {
                "bold": "bids-example/sub-{subject}/func/sub-{subject}_task-{task}_bold.nii.gz"
            },
            "input_zip_lists": {
                "bold": {
                    "subject": ["control01", "control01"],
                    "task": ["nback", "rest"]
                }
            },
            "input_lists": {
                "bold": {
                    "subject": ["control01"],
                    "task": ["nback", "rest"]
                }
            },
            "input_wildcards": {
                "bold": {
                    "subject": "{subject}",
                    "task": "{task}"
                }
            },
            "subjects": ["subject01"],
            "sessions": [],
            "subj_wildcards": {"subject": "{subject}"}
        }
    """  # noqa

    search_terms = _generate_search_terms(participant_label, exclude_participant_label)

    if os.path.exists(bids_dir):
        # generate inputs based on config
        layout = BIDSLayout(
            bids_dir,
            derivatives=derivatives,
            validate=False,
            indexer=BIDSLayoutIndexer(validate=False, index_metadata=False),
        )
    else:
        _logger.info(
            "bids_dir does not exist, skipping PyBIDS and using "
            "custom file paths only"
        )
        layout = None

    # this will populate input_path, input_lists, input_zip_lists, and
    # input_wildcards
    inputs_config_dict = _get_lists_from_bids(
        bids_layout=layout,
        pybids_inputs=pybids_inputs,
        limit_to=limit_to,
        **search_terms,
    )

    if layout is None:
        # if no layout, then use subjects/sessions from --path vars
        subjects = []
        sessions = []
        for input_type in inputs_config_dict["input_lists"]:

            subj_set = set(inputs_config_dict["input_lists"][input_type]["subject"])

            # filter the list of subjects with participant_label
            if participant_label is not None:
                subj_set = set.intersection(subj_set, set(participant_label))

            # TODO: need to also remove subjects based on exclude_participant_label

            # replace with filtered list
            inputs_config_dict["input_lists"][input_type]["subject"] = list(subj_set)

            # add to set of subjects from all input_types
            subjects.append(subj_set)

            if "session" in (inputs_config_dict["input_lists"][input_type].keys()):
                sessions.append(
                    {inputs_config_dict["input_lists"][input_type]["session"]}
                )
            else:
                sessions.append(set([]))

        # take set intersection of all input types
        inputs_config_dict["subjects"] = list(set.intersection(*subjects))
        inputs_config_dict["sessions"] = list(set.intersection(*sessions))

    else:
        # populate subjects, sessions and subj_wildcards in the config
        inputs_config_dict["subjects"] = layout.get_subjects(**search_terms)
        inputs_config_dict["sessions"] = layout.get_sessions(**search_terms)

    if len(inputs_config_dict["sessions"]) == 0:
        inputs_config_dict["subj_wildcards"] = {"subject": "{subject}"}
    else:
        inputs_config_dict["subj_wildcards"] = {
            "subject": "{subject}",
            "session": "{session}",
        }

    return inputs_config_dict


def _process_path_override(input_path):
    """Glob wildcards from a path override and arrange into a zip list of
    matches, list of matches, and Snakemake wildcard for each wildcard.
    """
    wildcards = glob_wildcards(input_path)
    wildcard_names = list(wildcards._fields)

    if len(wildcard_names) == 0:
        _logger.warning("No wildcards defined in %s", input_path)

    input_wildcards = {}
    input_zip_lists = {}
    input_lists = {}

    for i, wildcard in enumerate(wildcard_names):
        input_zip_lists[wildcard] = wildcards[i]
        input_lists[wildcard] = list(set(wildcards[i]))
        input_wildcards[wildcard] = f"{{{wildcard}}}"
        if len(wildcards[i]) == 0:
            _logger.error("No matching files for %s", input_path)

    return input_zip_lists, input_lists, input_wildcards


def _process_layout_wildcard(path, wildcard_name):
    """Convert an absolute BIDS path to the same path with the given tag
    replaced by a wildcard.

    Parameters
    ----------
    path : str
        Absolute BIDS path
        (e.g. "root/sub-01/ses-01/sub-01_ses-01_T1w.nii.gz")
    wildcard_name : str
        BIDS entity to replace with a wildcard. (e.g. "subject")

    Returns
    -------
    path : str
        Original path with the original entity replaced with a wildcard.
        (e.g. "root/sub-{subject}/ses-01/sub-{subject}_ses-01_T1w.nii.gz")
    match : str
        Matched BIDS entity in the input path (e.g. "01")
    out_name : str
        Name of the applied wildcard (e.g. "subject")
    """
    bids_tags = _read_bids_tags()
    tag = bids_tags[wildcard_name] if wildcard_name in bids_tags else wildcard_name

    # this changes e.g. sub-001 to sub-{subject} in the path
    # (so snakemake can use the wildcards)
    # HACK FIX FOR acq vs acquisition etc -- should
    # eventually update the bids() function to also use
    # bids_tags.json, where e.g. acquisition -> acq is
    # defined.. -- then, can use wildcard_name instead
    # of out_name..
    if wildcard_name not in ["subject", "session"]:
        out_name = tag
    else:
        out_name = wildcard_name

    if wildcard_name == "suffix":
        # capture suffix
        matching_pattern = ".*_([a-zA-Z0-9]+).*$"
        # capture before and after suffix
        replace_pattern = "(.*_)[a-zA-Z0-9]+(.*)$"
        # replace with before, {suffix}, after
        replace = f"\\1{{{out_name}}}\\2"
        match = re.search(matching_pattern, path)
        path = re.sub(replace_pattern, replace, path)

    else:
        pattern = f"{tag}-([a-zA-Z0-9]+)"
        replace = f"{tag}-{{{out_name}}}"

        match = re.search(pattern, path)
        path = re.sub(pattern, replace, path)

    # update the path with the {wildcards} -- uses the
    # value from the string (not from the pybids
    # entities), since that has issues with integer
    # formatting (e.g. for run=01)

    return path, match[1], out_name


def _get_lists_from_bids(bids_layout, pybids_inputs, limit_to=None, **filters):
    """Grabs files using pybids and creates snakemake-friendly lists

    Parameters
    ----------
    bids_layout : BIDSLayout
        Layout from pybids for accessing the BIDS dataset to grab paths

    pybids_inputs : dict
        Dictionary indexed by modality name, specifying the filters and
        wildcards for each pybids input.

    limit_to : list, optional
        List of inputs to skip, this used by snakebids to exclude modalities
        based on cmd-line args

    filters : dict, optional
        Pybids filters to apply globally to all inputs.

    Returns
    -------
    dict of dicts:
        There is one entry in the dictionary for each modality described by
        ``pybids_inputs``. Each of these entries contains the following items:

        * ``"input_path"``: String with a wildcard-filled path that matches
          the images for this modality.
        * ``"input_zip_lists"``: Dictionary where each key is a wildcard
          entity and each value is a list of the values found for that
          entity. Each of these lists has length equal to the number of
          images matched for this modality, so they can be zipped together to
          get a list of the wildcard values for each image.
        * ``"input_lists"``: Dictionary where each key is a wildcard entity
          and each value is a list of the unique values found for that
          entity. These lists may not be the same length.
        * ``"input_wildcards"``: Dictionary where each key is the name of a
          wildcard entity, and each value is the Snakemake wildcard used for
          that entity.
    """

    out_dict = dict(
        {
            "input_path": {},
            "input_zip_lists": {},
            "input_lists": {},
            "input_wildcards": {},
        }
    )

    if limit_to is None:
        limit_to = pybids_inputs.keys()

    for input_name in limit_to:
        _logger.debug("Grabbing inputs for %s...", input_name)
        input_path = ""
        input_wildcards = {}
        input_zip_lists = {}
        input_lists = {}
        if "custom_path" in pybids_inputs[input_name].keys():
            # a custom path was specified for this input, skip pybids:
            # get input_wildcards by parsing path for {} entries (using a set
            # to get unique only)
            # get input_zip_lists by using glob_wildcards (but need to modify
            # to deal with multiple wildcards

            input_path = pybids_inputs[input_name]["custom_path"]
            (
                input_zip_lists,
                input_lists,
                input_wildcards,
            ) = _process_path_override(input_path)
        else:
            paths = set()
            for img in bids_layout.get(
                **pybids_inputs[input_name]["filters"], **filters
            ):
                input_path = img.path
                for wildcard_name in pybids_inputs[input_name]["wildcards"]:

                    if wildcard_name not in img.get_entities():
                        continue
                    _logger.debug(
                        "Wildcard %s found entities for %s",
                        wildcard_name,
                        img.path,
                    )
                    (
                        input_path,
                        input_list,
                        out_name,
                    ) = _process_layout_wildcard(input_path, wildcard_name)
                    if out_name not in input_zip_lists:
                        input_zip_lists[out_name] = []
                        input_lists[out_name] = set()
                        input_wildcards[out_name] = {}
                    input_zip_lists[out_name].append(input_list)
                    input_lists[out_name].add(input_list)
                    input_wildcards[out_name] = f"{{{out_name}}}"
                paths.add(input_path)

            # now, check to see if unique
            if len(paths) == 0:
                _logger.warning("No images found for %s", input_name)
                continue
            if len(paths) > 1:
                _logger.warning(
                    "More than one snakemake filename for %s, taking the "
                    "first. To correct this, use the --filter_%s option to "
                    "narrow the search. Found filenames: %s",
                    input_name,
                    input_name,
                    paths,
                )

            input_path = list(paths)[0]

            # convert sets to lists
            input_lists = {key: list(val) for key, val in input_lists.items()}

        out_dict["input_path"][input_name] = input_path
        out_dict["input_zip_lists"][input_name] = input_zip_lists
        out_dict["input_lists"][input_name] = input_lists
        out_dict["input_wildcards"][input_name] = input_wildcards

    return out_dict


def get_wildcard_constraints(image_types):
    """Return a wildcard_constraints dict for snakemake to use, containing
    all the wildcards that are in the dynamically grabbed inputs

    Parameters
    ----------
    image_types : dict

    Returns
    -------
        Dict containing wildcard constraints for all wildcards in the
        inputs, with typical bids naming constraints, ie letters and numbers
        ``[a-zA-Z0-9]+``.
    """
    bids_constraints = "[a-zA-Z0-9]+"
    return {
        entity: bids_constraints
        for imgtype in image_types.keys()
        for entity in image_types[imgtype].keys()
    }


def write_derivative_json(snakemake, **kwargs):
    """Snakemake function to read a json file, and write to a new one,
    adding BIDS derivatives fields for Sources and Parameters.

    Parameters
    ----------
    snakemake : struct Snakemake
        structure passed to snakemake python scripts, containing input,
        output, params, config ...
        This function requires input.json and output.json to be defined, as
        it will read and write json files
    """

    with open(snakemake.input.json, "r", encoding="utf-8") as input_json:
        sidecar = json.load(input_json)

    sidecar.update(
        {
            "Sources": [snakemake.input],
            "Parameters": snakemake.params,
            **kwargs,
        }
    )

    with open(snakemake.output.json, "w", encoding="utf-8") as outfile:
        json.dump(sidecar, outfile, indent=4)



class ConfigError(Exception):
    """Exception raised for errors with the Snakebids config."""

    def __init__(self, msg):
        self.msg = msg
        super().__init__()


class RunError(Exception):
    """Exception raised for errors in generating and running the snakemake workflow."""

    def __init__(self, msg: str, *args: object) -> None:
        super().__init__(*args)
        self.msg = msg



"""Script to generate a Snakebids project."""

import os

from cookiecutter.main import cookiecutter

import snakebids


def main():
    """Invoke Cookiecutter on the Snakebids project template."""
    cookiecutter(os.path.join(snakebids.__path__[0], "../project_template"))



"""Tools to generate a Snakemake-based BIDS app."""

import argparse
import logging
import os
import pathlib
import re
import subprocess
import sys

import snakemake
from colorama import Fore
from snakemake.io import load_configfile

from snakebids.exceptions import ConfigError, RunError
from snakebids.utils.output import (
    prepare_output,
    retrofit_output,
    write_config_file,
    write_output_mode,
)

# We define Path here in addition to pathlib to put both variables in globals()
# This way, users specifying a path type in their config.yaml can indicate
# either Path or pathlib.Path
Path = pathlib.Path

logger = logging.Logger(__name__)


# pylint: disable=too-few-public-methods,
class KeyValue(argparse.Action):
    """Class for accepting key=value pairs in argparse"""

    # Constructor calling
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, {})

        for value in values:
            # split it into key and value
            key, value = value.split("=")
            # assign into dictionary
            getattr(namespace, self.dest)[key] = value


# pylint: disable=too-few-public-methods
class SnakemakeHelpAction(argparse.Action):
    """Class for printing snakemake usage in argparse"""

    def __call__(self, parser, namespace, values, option_string=None):
        run("snakemake -h")
        sys.exit(0)


def run(command, env=None):
    """Helper function for running a system command while merging
    stderr/stdout to stdout.

    Parameters
    ----------
    command : list of str
        command to run

    env : dict, optional
        environment variable to set before running the command
    """
    if env is None:
        env = {}

    merged_env = os.environ
    merged_env.update(env)
    # pylint: disable=consider-using-with
    process = subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        shell=True,
        env=merged_env,
    )
    while True:
        line = process.stdout.readline()
        line = str(line, "utf-8")[:-1]
        print(line)
        if line == "" and process.poll() is not None:
            break
    if process.returncode != 0:
        raise Exception(f"Non zero return code: {process.returncode}")


def resolve_path(path_candidate):
    """Helper function to resolve any paths or list
    of paths it's passed. Otherwise, returns the argument
    unchanged.

    Parameters
    ----------
    command : list, os.Pathlike, object
        command to run

    Returns
    -------
    list, os.Pathlike, object
        If os.Pathlike or list  of os.Pathlike, the same paths resolved.
        Otherwise, the argument unchanged.
    """
    if isinstance(path_candidate, list):
        return [resolve_path(p) for p in path_candidate]

    if isinstance(path_candidate, os.PathLike):
        return path_candidate.resolve()

    return path_candidate


SNAKEFILE_CHOICES = [
    "Snakefile",
    "snakefile",
    "workflow/Snakefile",
    "workflow/snakefile",
]


CONFIGFILE_CHOICES = [
    "config/snakebids.yml",
    "config/snakebids.yaml",
    "config/snakebids.json",
    "snakebids.yml",
    "snakebids.yaml",
    "snakebids.json",
    "config.yml",
    "config.yaml",
    "config.json",
    "config/config.json",
    "config/config.yml",
    "config/config.yaml",
]


# pylint: disable=too-few-public-methods, too-many-instance-attributes
class SnakeBidsApp:
    """Snakebids app with config and arguments.

    Parameters
    ----------
    snakemake_dir : str
        Root directory of the snakebids app, containing the config file and
        workflow files.
    skip_parse_args : bool, optional
        If true, the Snakebids app will not attempt to parse input arguments,
        and will only handle the config file.
    out_configfile : str
        Path to the updated configfile (YAML or JSON), relative to the output
        working directory. This should be the same as the `configfile: ` used
        in your workflow. (default: 'config/snakebids.yml')

    Attributes
    ----------
    config : dict
        Contains all the configuration variables parsed from the config file
        and generated during the initialization of the SnakeBidsApp.
    parser_include_snakemake : ArgumentParser
        Parser including the generic Snakemake parser as a parent. This will
        contain all arguments a Snakemake app can receive.
    parser : ArgumentParser
        Parser including only the arguments specific to this Snakebids app, as
        specified in the config file.
    snakefile : str
        Absolute path to the input Snakefile
            join(snakemake_dir, snakefile_path)
    configfile_path : str
        Relative path to config file (relative to snakemake_dir)
    updated_config : str
        Absolute path to the updated config file to write.
    workflow_mode : bool
        If true, produces output in workflow mode, otherwise snakemake mode.
    force : bool
        If false, prevents conversion of output from workflow mode to bidsapp mode.
    outputdir : Path
        Path to outputdir specified by user
    retrofit : bool
        If true, run_snakemake will attempt to convert legacy output format into bidsapp
        format.
    """

    outputdir: Path
    snakemake_dir: Path

    def __init__(self, snakemake_dir, skip_parse_args=False):
        # input argument is the dir where snakemake would be run
        # we use this to locate the config file, and snakefile adding them to
        # generated_config and also add the snakemake_dir to the
        # generated_config, so the workflow can use it to source files from
        # it (e.g. atlases etc..)

        self.snakemake_dir = Path(snakemake_dir).resolve()

        # look for snakebids.yml in the snakemake_dir, quit if not found
        self.configfile_path = ""
        for path in CONFIGFILE_CHOICES:
            if (self.snakemake_dir / path).exists():
                self.configfile_path = path
                break
        if not self.configfile_path:
            raise ConfigError(
                f"Error: no config file found, tried {', '.join(CONFIGFILE_CHOICES)}."
            )

        # look for snakefile in the snakemake_dir, quit if not found
        self.snakefile = None
        for snakefile_path in SNAKEFILE_CHOICES:
            if (self.snakemake_dir / snakefile_path).exists():
                self.snakefile = Path(snakemake_dir, snakefile_path)
                break
        if self.snakefile is None:
            raise ConfigError(
                f"Error: no Snakefile found, tried {', '.join(SNAKEFILE_CHOICES)}."
            )

        self.config = load_configfile(Path(snakemake_dir, self.configfile_path))

        if self.config.get("debug", False):
            logging.basicConfig(level=logging.DEBUG)

        # add path to snakefile to the config -- so workflows can grab files
        # relative to the snakefile folder
        self.config["snakemake_dir"] = snakemake_dir
        self.config["snakefile"] = self.snakefile

        self.parser_include_snakemake = self._create_parser(include_snakemake=True)
        self.parser = self._create_parser()

        if not skip_parse_args:
            self._parse_args()

    # pylint: disable=too-many-locals
    def _create_parser(self, include_snakemake=False):
        """Create a parser with snakemake parser as parent solely for
        displaying help and checking conflicts, but then for actual parsing
        use snakebids parser to parse known args, then pass remaining to
        snakemake.
        """

        if include_snakemake:
            # get snakemake parser
            smk_parser = snakemake.get_argument_parser()

            # create parser
            parser = argparse.ArgumentParser(
                description="Snakebids helps build BIDS Apps with Snakemake",
                add_help=False,
                parents=[smk_parser],
            )
        else:
            parser = argparse.ArgumentParser(
                description="Snakebids helps build BIDS Apps with Snakemake"
            )

        parser.add_argument(
            "--workflow-mode",
            "--workflow_mode",
            "-W",
            action="store_true",
            help=(
                "Run Snakebids in Workflow mode. This will cause the entire Snakebids "
                "app, except for the results folder, to be copied into your "
                "output_dir. Snakemake will be run in this new child app, and results "
                "will be put in output_dir/results."
            ),
        )

        # We use -x as the alias because both -f and -F are taken by snakemake
        parser.add_argument(
            "--force-conversion",
            "--force_conversion",
            "-x",
            action="store_true",
            help=(
                "Force snakebids to convert a workflow output to a bidsapp output. "
                "conversion will delete all the files in the workflow snakebids app."
            ),
        )

        parser.add_argument(
            "--retrofit",
            action="store_true",
            help=(
                "Convert a legacy Snakebids output (Snakebids 0.3.x or lower) into "
                "bidsapp mode. This will delete any config files currently in the "
                "output."
            ),
        )

        # add option for printing out snakemake usage
        parser.add_argument(
            "--help-snakemake",
            "--help_snakemake",
            nargs=0,
            action=SnakemakeHelpAction,
            help=(
                "Options to Snakemake can also be passed directly at the "
                "command-line, use this to print Snakemake usage"
            ),
        )

        # create parser group for app options
        app_group = parser.add_argument_group("SNAKEBIDS", "Options for snakebids app")

        # update the parser with config options
        for name, parse_args in self.config["parse_args"].items():
            # Convert type annotations from strings to class types
            # We first check that the type annotation is, in fact,
            # a str to allow the edge case where it's already
            # been converted
            if "type" in parse_args and isinstance(parse_args["type"], str):
                try:
                    parse_args["type"] = globals()[parse_args["type"]]
                except KeyError as err:
                    raise TypeError(
                        f"{parse_args['type']} is not available "
                        + f"as a type for {name}"
                    ) from err

            if re.match(r"^--", name):
                name_part = re.match(r"^--(.+)$", name).group(1)
                names = (
                    name,
                    re.sub(r"\_", "-", name),
                    f"--{re.sub(r'-', '_', name_part)}",
                )
            else:
                names = (name,)
            app_group.add_argument(*names, **parse_args)

        # general parser for
        # --filter_{input_type} {key1}={value1} {key2}={value2}...
        # create filter parsers, one for each input_type
        filter_opts = parser.add_argument_group(
            "BIDS FILTERS",
            "Filters to customize PyBIDS get() as key=value pairs",
        )

        for input_type in self.config["pybids_inputs"].keys():
            argnames = (f"--filter_{input_type}", f"--filter-{input_type}")
            arglist_default = [
                f"{key}={value}"
                for (key, value) in self.config["pybids_inputs"][input_type][
                    "filters"
                ].items()
            ]
            arglist_default_string = " ".join(arglist_default)

            filter_opts.add_argument(
                *argnames,
                nargs="+",
                action=KeyValue,
                help=f"(default: {arglist_default_string})",
            )

        # general parser for
        # --wildcards_{input_type} {wildcard1} {wildcard2} ...
        # create wildcards parsers, one for each input_type
        wildcards_opts = parser.add_argument_group(
            "INPUT WILDCARDS",
            "File path entities to use as wildcards in snakemake",
        )

        for input_type in self.config["pybids_inputs"].keys():
            argnames = (f"--wildcards-{input_type}", f"--wildcards_{input_type}")
            arglist_default = [
                f"{wc}" for wc in self.config["pybids_inputs"][input_type]["wildcards"]
            ]
            arglist_default_string = " ".join(arglist_default)

            wildcards_opts.add_argument(
                *argnames,
                nargs="+",
                help=f"(default: {arglist_default_string})",
            )

        override_opts = parser.add_argument_group(
            "PATH OVERRIDE",
            (
                "Options for overriding BIDS by specifying absolute paths "
                "that include wildcards, e.g.: "
                "/path/to/my_data/{subject}/t1.nii.gz"
            ),
        )

        # create path override parser
        for input_type in self.config["pybids_inputs"].keys():
            argnames = (f"--path-{input_type}", f"--path_{input_type}")
            override_opts.add_argument(*argnames, default=None)

        return parser

    def _parse_args(self):

        # use snakebids parser to parse the known arguments
        # will pass the rest of args when running snakemake
        all_args = self.parser.parse_known_args()

        args = all_args[0]
        snakemake_args = all_args[1]
        # resolve all path items to get absolute paths
        args.__dict__ = {k: resolve_path(v) for k, v in args.__dict__.items()}

        self.workflow_mode = args.workflow_mode
        self.force = args.force_conversion
        self.outputdir = Path(args.output_dir).resolve()
        self.retrofit = args.retrofit

        # add snakebids arguments to config
        self.config.update(args.__dict__)

        # add snakemake arguments to config
        self.config.update({"snakemake_args": snakemake_args})

        # argparse adds filter_{input_type} to the config
        # we want to update the pybids_inputs dict with this, then remove the
        # filter_{input_type} dict
        for input_type in self.config["pybids_inputs"].keys():
            arg_filter_dict = self.config[f"filter_{input_type}"]
            if arg_filter_dict is not None:
                self.config["pybids_inputs"][input_type]["filters"].update(
                    arg_filter_dict
                )
            del self.config[f"filter_{input_type}"]

        # add cmdline defined wildcards from the list:
        # wildcards_{input_type}
        for input_type in self.config["pybids_inputs"].keys():
            wildcards_list = self.config[f"wildcards_{input_type}"]
            if wildcards_list is not None:
                self.config["pybids_inputs"][input_type]["wildcards"] += wildcards_list
            del self.config[f"wildcards_{input_type}"]

        # add custom input paths to
        # config['pybids_inputs'][input_type]['custom_path']
        for input_type in self.config["pybids_inputs"].keys():
            custom_path = self.config[f"path_{input_type}"]
            if custom_path is not None:
                self.config["pybids_inputs"][input_type]["custom_path"] = Path(
                    custom_path
                ).resolve()
            del self.config[f"path_{input_type}"]

        # replace paths with realpaths
        self.config["bids_dir"] = Path(self.config["bids_dir"]).resolve()
        self.config["output_dir"] = Path(self.config["output_dir"]).resolve()

    def run_snakemake(self):
        """Run snakemake with that config.

        Workflow snakefile will read snakebids config, create inputs_config,
        and read that in.
        """

        if self.snakemake_dir == self.outputdir:
            write_output_mode(self.outputdir / ".snakebids", "workflow")
            mode = "workflow"
            force_config_overwrite = True
            if not self.workflow_mode:
                print(
                    f"{Fore.YELLOW}You specified your output to be in the snakebids "
                    "directory, so we're switching automatically to workflow mode!\n"
                    f"{Fore.RESET}You'll find your results in "
                    f"`{(self.snakemake_dir/'results').resolve()}`"
                )
        else:
            mode = "workflow" if self.workflow_mode else "bidsapp"
            force_config_overwrite = False
            if self.retrofit:
                for path in CONFIGFILE_CHOICES:
                    if (self.outputdir / path).exists():
                        self.configfile_path = path
                if not retrofit_output(
                    self.outputdir,
                    # Find all config files in the outputdir
                    (
                        self.outputdir / p
                        for p in CONFIGFILE_CHOICES
                        if (self.outputdir / p).exists()
                    ),
                ):
                    print(f"{Fore.YELLOW}Exiting. No conversion performed.{Fore.RESET}")
                    sys.exit(1)

        try:
            root = prepare_output(self.snakemake_dir, self.outputdir, mode, self.force)
        except RunError as err:
            print(err.msg)
            sys.exit(1)

        if mode == "workflow":
            self.config["output_dir"] = str(root)
            self.config["root"] = "results"
            new_config_file = self.outputdir / self.configfile_path
        else:
            force_config_overwrite = True
            self.config["root"] = ""
            new_config_file = self.outputdir / self.configfile_path

        cwd = self.outputdir

        write_config_file(new_config_file, self.config, force_config_overwrite)

        # running the chosen participant level
        analysis_level = self.config["analysis_level"]

        # run snakemake (passing any leftover args from argparse)
        # Filter any blank strings before submitting
        snakemake_argv = [
            *filter(
                None,
                [
                    "--snakefile",
                    str(self.snakefile),
                    "--directory",
                    str(cwd),
                    "--configfile",
                    str(new_config_file.resolve()),
                    *self.config["snakemake_args"],
                    *self.config["targets_by_analysis_level"][analysis_level],
                ],
            )
        ]

        snakemake.main(snakemake_argv)



__submodules__ = ["bids"]

# <AUTOGEN_INIT>
from snakebids.bids import (
    bids,
    filter_list,
    generate_inputs,
    get_filtered_ziplist_index,
    get_wildcard_constraints,
    print_boilerplate,
    write_derivative_json,
)

__all__ = [
    "bids",
    "filter_list",
    "generate_inputs",
    "get_filtered_ziplist_index",
    "get_wildcard_constraints",
    "print_boilerplate",
    "write_derivative_json",
]

# </AUTOGEN_INIT>



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['snakebids', 'snakebids.utils']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.3.1,<6.0.0',
 'colorama>=0.4.4,<0.5.0',
 'cookiecutter>=1.7.2,<2.0.0',
 'progress>=1.6,<2.0',
 'pybids>=0.14.0,<0.15.0',
 'snakemake>=5.28.0',
 'typing-extensions>=3.10.0,<4.0.0']

entry_points = \
{'console_scripts': ['snakebids-create = snakebids.template:main']}

setup_kwargs = {
    'name': 'snakebids',
    'version': '0.4.0',
    'description': 'BIDS integration into snakemake workflows',
    'long_description': "\nsnakebids\n=========\n.. image:: https://readthedocs.org/projects/snakebids/badge/?version=latest\n  :target: https://snakebids.readthedocs.io/en/latest/?badge=latest\n  :alt: Documentation Status\n\nSnakemake + BIDS\n\nThis package allows you to build BIDS Apps using Snakemake. It offers:\n\n\n* Flexible data grabbing with PyBIDS, configurable solely by config file entries\n* Helper function for creating BIDS paths inside Snakemake workflows/rules\n* Command-line invocation of snakemake workflows with BIDS App compliance\n* Configurable argument parsing specified using the Snakemake workflow config\n* Execution either as command-line BIDS apps or via snakemake executable\n\nContributing\n============\n\nClone the git repository. Snakebids dependencies are managed with Poetry, which you'll need installed on your machine. You can find instructions on the `poetry website <https://python-poetry.org/docs/master/#installation>`_. Then, setup the development environment with the following commands::\n\n  poetry install\n  poetry run poe setup\n\nSnakebids uses `poethepoet <https://github.com/nat-n/poethepoet>`_ as a task runner. You can see what commands are available by running::\n\n    poetry run poe\n\nIf you wish, you can also run ``poe [[command]]`` directly by installing ``poethepoet`` on your system. Follow the install instructions at the link above.\n\nTests are done with ``pytest`` and can be run via::\n\n  poetry run pytest\n\nSnakebids uses pre-commit hooks (installed via the ``poe setup`` command above) to lint and format code (we use `black <https://github.com/psf/black>`_, `isort <https://github.com/PyCQA/isort>`_, `pylint <https://pylint.org/>`_ and `flake8 <https://flake8.pycqa.org/en/latest/>`_). By default, these hooks are run on every commit. Please be sure they all pass before making a PR.\n",
    'author': 'Ali Khan',
    'author_email': 'alik@robarts.ca',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/akhanf/snakebids',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



""" Minor modification to allow glob_wildcards() to have multiple occurence of
same wildcard """

import collections
import os
import re
from itertools import chain


def regex(filepattern):
    """Build Snakebids regex based on the given file pattern."""
    regex_list = []
    last = 0
    wildcards = set()
    for match in _wildcard_regex.finditer(filepattern):
        regex_list.append(re.escape(filepattern[last : match.start()]))
        wildcard = match.group("name")
        if wildcard in wildcards:
            if match.group("constraint"):
                raise ValueError(
                    "Constraint regex must be defined only in the first "
                    "occurence of the wildcard in a string."
                )
            regex_list.append(f"(?P={wildcard})")
        else:
            wildcards.add(wildcard)
            constraint = (
                match.group("constraint") if match.group("constraint") else ".+"
            )
            regex_list.append(f"(?P<{wildcard}>{constraint})")

        last = match.end()
    regex_list.append(re.escape(filepattern[last:]))
    regex_list.append("$")  # ensure that the match spans the whole file
    return "".join(regex_list)


_wildcard_regex = re.compile(
    r"""
    \{
        (?=(   # This lookahead assertion emulates an 'atomic group'
               # which is required for performance
            \s*(?P<name>\w+)                    # wildcard name
            (\s*,\s*
                (?P<constraint>                 # an optional constraint
                    ([^{}]+ | \{\d+(,\d+)?\})*  # allow curly braces to nest
                )                               # ...  as in '{w,a{3,5}}'
            )?\s*
        ))\1
    \}
    """,
    re.VERBOSE,
)


def glob_wildcards(pattern, files=None, followlinks=False):
    """Glob the values of wildcards by matching a pattern to the filesystem.

    Parameters
    ----------
    pattern : str
        Path including wildcards to glob on the filesystem.
    files : list of str, optional
        Files from which to glob wildcards. If None (default), the directory
        corresponding to the first wildcard in the pattern is walked, and
        wildcards are globbed from all files.
    followlinks : bool, optional
        Whether to follow links when globbing wildcards. Default: False.

    Returns
    -------
    namedtuple
        "Wildcards" named tuple where each field name is the name of a
        wildcard and the value of each field is a list of values for the
        corresponding wildcard.
    """
    pattern = os.path.normpath(pattern)
    first_wildcard = re.search("{[^{]", pattern)
    dirname = (
        os.path.dirname(pattern[: first_wildcard.start()])
        if first_wildcard
        else os.path.dirname(pattern)
    )
    if not dirname:
        dirname = "."

    names = [match.group("name") for match in _wildcard_regex.finditer(pattern)]

    # remove duplicates while preserving ordering
    names = list(collections.OrderedDict.fromkeys(names))

    # pylint: disable-msg=invalid-name
    Wildcards = collections.namedtuple("Wildcards", names)

    wildcards = Wildcards(*[[] for _ in names])

    pattern = re.compile(regex(pattern))

    if files is None:
        files = (
            os.path.normpath(os.path.join(dirpath, f))
            for dirpath, dirnames, filenames in os.walk(
                dirname, followlinks=followlinks
            )
            for f in chain(filenames, dirnames)
        )

    for f in files:
        match = re.match(pattern, f)
        if match:
            for name, value in match.groupdict().items():
                getattr(wildcards, name).append(value)
    return wildcards


def update_wildcard_constraints(
    pattern, wildcard_constraints, global_wildcard_constraints
):
    """Update wildcard constraints.

    Parameters
    ----------
    pattern : str
        Pattern on which to update constraints.
    wildcard_constraints : dict
        Dictionary of wildcard:constraint key-value pairs.
    global_wildcard_constraints : dict
        Dictionary of wildcard:constraint key-value pairs.
    """

    def replace_constraint(match):
        name = match.group("name")
        constraint = match.group("constraint")
        newconstraint = wildcard_constraints.get(
            name, global_wildcard_constraints.get(name)
        )
        if name in examined_names:
            return match.group(0)
        examined_names.add(name)
        # Don't override if constraint already set
        if constraint is not None:
            return match.group(0)
        # Only update if a new constraint has actually been set
        if newconstraint is not None:
            return f"{{{name},{newconstraint}}}"
        return match.group(0)

    examined_names = set()
    updated = _wildcard_regex.sub(replace_constraint, pattern)

    return updated



import setuptools
import json

with open("README.rst", "r") as fh:
    long_description = fh.read()

with open('pipeline_description.json', 'r') as fh:
    pipeline = json.load(fh)
    name = pipeline['GeneratedBy'][0]['Name']
    description = pipeline['Name']
    version = pipeline['GeneratedBy'][0]['Version']
    url = pipeline['GeneratedBy'][0]['CodeURL']
    author = pipeline['GeneratedBy'][0]['Author']
    author_email = pipeline['GeneratedBy'][0]['AuthorEmail']
 
setuptools.setup(
    name=name,
    version=version,
    author=author,
    author_email=author_email,
    description=description,
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url=url,
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={'console_scripts': [
        '{{cookiecutter.app_name}}={{cookiecutter.app_name}}.run:main'
    ]},
    install_requires=[
        "snakebids>={{cookiecutter.snakebids_version}}",
        "snakemake"
    ],
    python_requires='>=3.7'
)
